@@ -17,6 +17,18 @@
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
+#ifdef HAVE_CONFIG_H
+#  include "config.h"
+/* We need to undefine them because my_config.h defines them. :< */
+#  undef VERSION
+#  undef PACKAGE
+#  undef PACKAGE_BUGREPORT
+#  undef PACKAGE_NAME
+#  undef PACKAGE_STRING
+#  undef PACKAGE_TARNAME
+#  undef PACKAGE_VERSION
+#endif
+
 #ifdef USE_PRAGMA_IMPLEMENTATION
 #pragma implementation
 #endif
@@ -92,6 +104,19 @@ grn_logger_info mrn_logger_info = {
   NULL
 };
 
+/* global hashes and mutexes */
+HASH mrn_allocated_thds;
+pthread_mutex_t mrn_allocated_thds_mutex;
+uchar *mrn_allocated_thds_get_key(
+  THD *thd,
+  size_t *length,
+  my_bool not_used __attribute__ ((unused))
+) {
+  DBUG_ENTER("mrn_allocated_thds_get_key");
+  *length = sizeof(THD *);
+  DBUG_RETURN((uchar*) thd);
+}
+
 /* system functions */
 
 static void mrn_create_status()
@@ -183,7 +208,7 @@ my_bool last_insert_grn_id_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
 int last_insert_grn_id(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error)
 {
   THD *thd = current_thd;
-  st_mrn_slot_data *slot_data = (st_mrn_slot_data*) thd_get_ha_data(thd, mrn_hton_ptr);
+  st_mrn_slot_data *slot_data = (st_mrn_slot_data*) *thd_ha_data(thd, mrn_hton_ptr);
   if (slot_data == NULL) {
     return 0;
   }
@@ -323,7 +348,7 @@ int mrn_init(void *p)
   ctx = grn_ctx_open(0);
 
   if (pthread_mutex_init(&mrn_log_mutex, NULL) != 0) {
-    goto err;
+    goto err_log_mutex_init;
   }
   grn_logger_info_set(ctx, &mrn_logger_info);
   if (!(mrn_logfile = fopen(mrn_logfile_name, "a"))) {
@@ -351,13 +376,27 @@ int mrn_init(void *p)
 
   // init lock
   if ((pthread_mutex_init(&db_mutex, NULL) != 0)) {
-    goto err;
+    goto err_db_mutex_init;
+  }
+  if ((pthread_mutex_init(&mrn_allocated_thds_mutex, NULL) != 0)) {
+    goto err_allocated_thds_mutex_init;
+  }
+  if (my_hash_init(&mrn_allocated_thds, system_charset_info, 32, 0, 0,
+                   (my_hash_get_key) mrn_allocated_thds_get_key, 0, 0)) {
+    goto error_allocated_thds_hash_init;
   }
 
   grn_ctx_fin(ctx);
   return 0;
 
+error_allocated_thds_hash_init:
+  pthread_mutex_destroy(&mrn_allocated_thds_mutex);
+err_allocated_thds_mutex_init:
+  pthread_mutex_destroy(&db_mutex);
+err_db_mutex_init:
 err:
+  pthread_mutex_destroy(&mrn_log_mutex);
+err_log_mutex_init:
   grn_ctx_fin(ctx);
   grn_fin();
   return -1;
@@ -365,11 +404,26 @@ err:
 
 int mrn_deinit(void *p)
 {
+  THD *thd = current_thd, *tmp_thd;
   grn_ctx *ctx;
   ctx = grn_ctx_open(0);
 
   GRN_LOG(ctx, GRN_LOG_NOTICE, "%s deinit", MRN_PACKAGE_STRING);
 
+  if (thd && thd_sql_command(thd) == SQLCOM_UNINSTALL_PLUGIN) {
+    pthread_mutex_lock(&mrn_allocated_thds_mutex);
+    while ((tmp_thd = (THD *) my_hash_element(&mrn_allocated_thds, 0)))
+    {
+      void *slot_ptr = *thd_ha_data(tmp_thd, mrn_hton_ptr);
+      if (slot_ptr) free(slot_ptr);
+      *thd_ha_data(tmp_thd, mrn_hton_ptr) = (void *) NULL;
+      my_hash_delete(&mrn_allocated_thds, (uchar *) tmp_thd);
+    }
+    pthread_mutex_unlock(&mrn_allocated_thds_mutex);
+  }
+
+  my_hash_free(&mrn_allocated_thds);
+  pthread_mutex_destroy(&mrn_allocated_thds_mutex);
   pthread_mutex_destroy(&mrn_log_mutex);
   pthread_mutex_destroy(&db_mutex);
   grn_hash_close(ctx, mrn_hash);
@@ -412,11 +466,16 @@ void mrn_drop_db(handlerton *hton, char *path)
 
 int mrn_close_connection(handlerton *hton, THD *thd)
 {
-  void *p = thd_get_ha_data(thd, mrn_hton_ptr);
-  if (p) free(p);
-  thd_set_ha_data(thd, hton, NULL);
+  void *p = *thd_ha_data(thd, mrn_hton_ptr);
+  if (p) {
+    free(p);
+    *thd_ha_data(thd, mrn_hton_ptr) = (void *) NULL;
+    pthread_mutex_lock(&mrn_allocated_thds_mutex);
+    my_hash_delete(&mrn_allocated_thds, (uchar*) thd);
+    pthread_mutex_unlock(&mrn_allocated_thds_mutex);
+  }
   return 0;
-} 
+}
 
 bool mrn_flush_logs(handlerton *hton)
 {
@@ -816,7 +875,8 @@ ulonglong ha_mroonga_table_flags =
     HA_NO_AUTO_INCREMENT |
     HA_CAN_INSERT_DELAYED |
     HA_BINLOG_FLAGS |
-    HA_CAN_BIT_FIELD;
+    HA_CAN_BIT_FIELD |
+    HA_DUPLICATE_POS;
     //HA_HAS_RECORDS;
 
 ulonglong ha_mroonga::table_flags() const
@@ -1280,6 +1340,11 @@ int ha_mroonga::info(uint flag)
   DBUG_ENTER("ha_mroonga::info");
   ha_rows rows = grn_table_size(ctx, tbl);
   stats.records = rows;
+
+  if (flag & (HA_STATUS_ERRKEY | HA_STATUS_NO_LOCK)) {
+    errkey = dup_key;
+  }
+
   DBUG_RETURN(0);
 }
 
@@ -1417,9 +1482,10 @@ int ha_mroonga::write_row(uchar *buf)
 #ifndef DBUG_OFF
     dbug_tmp_restore_column_map(table->read_set, tmp_map);
 #endif
+    memcpy(dup_ref, &row_id, sizeof(grn_id));
+    dup_key = pkeynr;
     GRN_LOG(ctx, GRN_LOG_ERROR, "duplicated _id on insert");
-    my_message(ER_DUP_UNIQUE, "duplicated _id on insert", MYF(0));
-    DBUG_RETURN(ER_DUP_UNIQUE);
+    DBUG_RETURN(HA_ERR_FOUND_DUPP_KEY);
   }
 
   grn_obj colbuf;
@@ -1460,12 +1526,18 @@ int ha_mroonga::write_row(uchar *buf)
   grn_obj_unlink(ctx, &colbuf);
 
   // for UDF last_insert_grn_id()
-  st_mrn_slot_data *slot_data = (st_mrn_slot_data*) thd_get_ha_data(thd, mrn_hton_ptr);
+  st_mrn_slot_data *slot_data = (st_mrn_slot_data*) *thd_ha_data(thd, mrn_hton_ptr);
   if (slot_data == NULL) {
     slot_data = (st_mrn_slot_data*) malloc(sizeof(st_mrn_slot_data)); 
+    *thd_ha_data(thd, mrn_hton_ptr) = (void *) slot_data;
+    pthread_mutex_lock(&mrn_allocated_thds_mutex);
+    if (my_hash_insert(&mrn_allocated_thds, (uchar*) thd))
+    {
+      DBUG_RETURN(HA_ERR_OUT_OF_MEM);
+    }
+    pthread_mutex_unlock(&mrn_allocated_thds_mutex);
   }
   slot_data->last_insert_rid = row_id;
-  thd_set_ha_data(thd, mrn_hton_ptr, slot_data);
 
   DBUG_RETURN(0);
 }
@@ -2471,7 +2543,7 @@ void ha_mroonga::store_fields_from_primary_table(uchar *buf, grn_id rid)
 
 int ha_mroonga::reset()
 {
-  DBUG_ENTER("ha_mroonga::reset()");
+  DBUG_ENTER("ha_mroonga::reset");
   if (sort_keys != NULL) {
     free(sort_keys);
     sort_keys = NULL;
